"""Generate markdown changelog of Xsuite.

This script requires `gh` to be installed.
On macOS this can be obtained with `brew install gh`.
Afterwards it is necessary to authenticate with GitHub:

$ gh auth login

Then, a changelog can be generated with:

$ python make_changelog.py --start [date] --end [date]

Run `python run_on_gh.py --help` for usage information.
"""
from dateutil.parser import isoparse

import click
import json
import subprocess


PACKAGES = ['xdeps', 'xobjects', 'xpart', 'xtrack', 'xfields', 'xmask']


def fetch_package_log_page(package, page_size, page):
    command = (
        'gh', 'api',
        '-H', 'Accept: application/vnd.github+json',
        '-H', 'X-GitHub-Api-Version: 2022-11-28',
        f'/repos/xsuite/{package}/releases?per_page={page_size}&page={page}'
    )
    p = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    if p.stderr:
        raise RuntimeError(p.stderr)

    return json.loads(p.stdout)


def fetch_package_log(package, end):
    page = 1
    full_log = []
    while log := fetch_package_log_page(package, 100, page):
        last_publish_date = isoparse(log[-1]['published_at'])

        if end and last_publish_date > end:
            # Skip pages that are too new
            break

        full_log += log
        page += 1

    return full_log


def print_entry(entry):
    date = isoparse(entry["published_at"])
    print(f'# [{entry["name"]}]({entry["url"]})')
    print()
    print(f'Published: {date.strftime("%Y-%m-%d %H:%M")}')
    print()
    print(entry['body'])
    print()


@click.command()
@click.option('--start', '-s',
              type=click.DateTime(formats=["%Y-%m-%d"]))
@click.option('--end', '-e',
              type=click.DateTime(formats=["%Y-%m-%d"]))
def make_changelog(start, end):
    if start:
        start = start.astimezone()

    if end:
        end = end.astimezone()

    joint_log = []
    for pkg in PACKAGES:
        joint_log += fetch_package_log(pkg, end)

    joint_log = sorted(joint_log, key=lambda entry: entry['published_at'])

    for entry in joint_log:
        published_at = isoparse(entry['published_at'])

        if start and published_at < start:
            continue

        if end and published_at > end:
            continue

        print_entry(entry)


if __name__ == '__main__':
    make_changelog()



# copyright ############################### #
# This file is part of the Xsuite project.  #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import json
import os
from multiprocessing import Pool
from pathlib import Path
from pprint import pformat
from typing import Iterator, Optional, Tuple

import xsuite as xs
import xcoll as xc
from xtrack.prebuilt_kernel_definitions import XTRACK_ELEMENTS_INIT_DEFAULTS
from xfields.prebuilt_kernel_definitions import XFIELDS_ELEMENTS_INIT_DEFAULTS
from xcoll.prebuilt_kernel_definitions import XCOLL_ELEMENTS_INIT_DEFAULTS
import xfields as xf
import xobjects as xo
import xtrack as xt
from xsuite.kernel_definitions import kernel_definitions
from xtrack.general import _print

XSK_PREBUILT_KERNELS_LOCATION = Path(xs.__file__).parent / 'lib'

BEAM_ELEMENTS_INIT_DEFAULTS = XTRACK_ELEMENTS_INIT_DEFAULTS| XFIELDS_ELEMENTS_INIT_DEFAULTS \
                            | XCOLL_ELEMENTS_INIT_DEFAULTS


def get_element_class_by_name(name: str) -> type:
    extra_classes = (xt.MultiSetter, )

    element_classes = (xt.element_classes + xt.rng_classes + xt.monitor_classes
                       + xf.element_classes + xc.element_classes
                       + extra_classes)

    for cls in element_classes:
        if cls.__name__ == name:
            return cls

    raise ValueError(f'No element class with name {name} available.')


def save_kernel_metadata(
        module_name: str,
        config: dict,
        kernel_element_classes,
        location,
):
    location = Path(location)
    out_file = location / f'{module_name}.json'

    kernel_metadata = {
        'config': config.data,
        'classes': [cls._DressingClass.__name__ for cls in kernel_element_classes],
        'versions': {
            'xtrack': xt.__version__,
            'xfields': xf.__version__,
            'xcoll': xc.__version__,
            'xobjects': xo.__version__,
        }
    }

    with out_file.open('w') as fd:
        json.dump(kernel_metadata, fd, indent=4)


def enumerate_kernels(verbose=False) -> Iterator[Tuple[str, dict]]:
    """
    Iterate over the prebuilt kernels compatible with the current version of
    xsuite. The first element of the tuple is the name of the kernel module
    and the second is a dictionary with the kernel metadata.
    """
    for kernel_name, _ in kernel_definitions:
        metadata_file = XSK_PREBUILT_KERNELS_LOCATION / f'{kernel_name}.json'

        if not metadata_file.exists():
            continue

        with metadata_file.open('r') as fd:
            kernel_metadata = json.load(fd)

        needed_versions = kernel_metadata['versions']
        have_versions = {
            'xtrack': xt.__version__,
            'xfields': xf.__version__,
            'xcoll': xc.__version__,
            'xobjects': xo.__version__,
        }

        version_mismatch = False
        for package in needed_versions.keys():
            need = needed_versions[package]
            have = have_versions[package]
            if need == have:
                continue

            version_mismatch = True
            if verbose:
                _print(
                    f'Version mismatch for kernel `{kernel_name}`: needs '
                    f'{package}=={need}, but have {package}=={have}.'
                )

        if version_mismatch:
            continue

        yield metadata_file.stem, kernel_metadata


def get_suitable_kernel(
        config: dict,
        line_element_classes,
        verbose=False,
) -> Optional[Tuple[str, list]]:
    """
    Given a configuration and a list of element classes, return a tuple with
    the name of a suitable prebuilt kernel module together with the list of
    element classes that were used to build it. Set `verbose` to True, to
    obtain a justification of the choice (or lack thereof) on standard output.
    """
    env_var = os.environ.get("XSUITE_PREBUILT_KERNELS")
    if env_var and env_var == '0':
        if verbose:
            _print('Skipping the search for a suitable kernel, as the '
                   'environment variable XSUITE_PREBUILT_KERNELS == "0".')
        return

    if os.environ.get("XSUITE_VERBOSE", None) is not None:
        verbose = True

    requested_class_names = [
        cls._DressingClass.__name__ for cls in line_element_classes
    ]

    for module_name, kernel_metadata in enumerate_kernels(verbose=verbose):
        if verbose:
            _print(f"==> Considering the precompiled kernel `{module_name}`...")

        available_classes_names = kernel_metadata['classes']
        if kernel_metadata['config'] != config:
            if verbose:
                lhs = kernel_metadata['config']
                rhs = config
                config_diff = {kk: (lhs.get(kk), rhs.get(kk))
                               for kk in set(lhs.keys()) | set(rhs.keys())
                               if lhs.get(kk) != rhs.get(kk)}
                _print(f'The kernel `{module_name}` is unsuitable. Its config '
                      f'(left) and the requested one (right) differ at the '
                      f'following keys:\n'
                      f'{pformat(config_diff)}')
                _print(f'Skipping class compatibility check for `{module_name}`.')

            continue

        if verbose:
            _print(f'The kernel `{module_name}` has the right config.')

        if set(requested_class_names) <= set(available_classes_names):
            available_classes = [
                get_element_class_by_name(class_name)
                for class_name in available_classes_names
            ]
            if verbose:
                _print(f'Found suitable prebuilt kernel `{module_name}`.')
            return module_name, available_classes
        elif verbose:
            class_diff = set(requested_class_names) - set(available_classes_names)
            _print(f'The kernel `{module_name}` is unsuitable. It does not '
                  f'provide the following requested classes: '
                  f'{", ".join(class_diff)}.')

    if verbose:
        _print('==> No suitable precompiled kernel found.')


def regenerate_kernels(
        kernels=None,
        location=XSK_PREBUILT_KERNELS_LOCATION,
        n_threads=None,
):
    """
    Use the kernel definitions in the `kernel_definitions.py` file to
    regenerate kernel shared objects using the current version of xsuite.
    """
    if kernels is not None and (
    isinstance(kernels, str) or not hasattr(kernels, '__iter__')):
        kernels = [kernels]

    location = Path(location)

    # Delete existing kernels to avoid accidentally loading in existing C code
    clear_kernels(kernels=kernels, location=location)

    kernels_to_build = []
    for module_name, metadata in kernel_definitions:
        if kernels is not None and module_name not in kernels:
            continue
        kernels_to_build.append((module_name, metadata))

    thread_pool = Pool(processes=n_threads)
    results = []
    for idx, (module_name, metadata) in enumerate(kernels_to_build):
        args = (idx, len(kernels_to_build), location, metadata, module_name)
        result = thread_pool.apply_async(build_single_kernel, args=args)
        results.append(result)

    thread_pool.close()
    thread_pool.join()

    # Ensure no errors
    for result in results:
        result.get()

    _print(f'Built {len(kernels_to_build)} kernels.')


def build_single_kernel(idx, total, location, metadata, module_name):
    _print(f'[{idx + 1}/{total}] Building `{module_name}`...')

    config = metadata['config']
    element_classes = metadata['classes']
    extra_classes = metadata.get('extra_classes', [])

    elements = []
    for cls in element_classes:
        if cls.__name__ in BEAM_ELEMENTS_INIT_DEFAULTS:
            element = cls(**BEAM_ELEMENTS_INIT_DEFAULTS[cls.__name__])
        else:
            element = cls()
        elements.append(element)

    line = xt.Line(elements=elements)
    tracker = xt.Tracker(line=line, compile=False, _prebuilding_kernels=True)
    tracker.config.clear()
    tracker.config.update(config)
    tracker_classes = tracker._tracker_data_base.kernel_element_classes
    expected_classes = [getattr(el, '_XoStruct', el) for el in element_classes]
    extra_classes.extend(ee for ee in expected_classes if ee not in tracker_classes)

    # Get all kernels in the elements
    extra_kernels = {}
    extra_classes.append(xt.Particles)
    extra_classes = [getattr(el, '_XoStruct', el) for el in extra_classes]

    all_classes = tracker._tracker_data_base.kernel_element_classes + extra_classes

    for el in all_classes:
        extra_kernels.update(el._kernels)

    tracker._build_kernel(
        module_name=module_name,
        containing_dir=location,
        compile='force',
        extra_classes=extra_classes,
        extra_kernels=extra_kernels,
    )

    all_classes = [cls for cls in all_classes if cls.__name__ != 'ParticlesData']

    save_kernel_metadata(
        module_name=module_name,
        config=tracker.config,
        kernel_element_classes=all_classes,
        location=location,
    )


def clear_kernels(kernels=None, verbose=False, location=XSK_PREBUILT_KERNELS_LOCATION):
    if kernels is not None and (
            isinstance(kernels, str) or not hasattr(kernels, '__iter__')):
        kernels = [kernels]

    location = Path(location)

    for file in location.iterdir():
        if file.name.startswith('_'):
            continue
        if file.suffix not in ('.c', '.so', '.json'):
            continue
        if kernels is not None and file.stem.split('.')[0] not in kernels:
            continue
        file.unlink()

        if verbose:
            print(f'Removed `{file}`.')


if __name__ == '__main__':
    regenerate_kernels()




# copyright ################################# #
# This file is part of the Xsuite Package.    #
# Copyright (c) CERN, 2024.                   #
# ########################################### #

from importlib.metadata import version
__version__ = version(__name__)

from .prebuild_kernels import get_suitable_kernel, XSK_PREBUILT_KERNELS_LOCATION



"""Run Xsuite tests from a CLI using GitHub Actions.

This script requires `gh` to be installed.
On macOS this can be obtained with `brew install gh`.
Afterwards it is necessary to authenticate with GitHub:

$ gh auth login

and select `xsuite/xsuite` as the default repository:

$ gh repo set-default xsuite/xsuite

Run `python run_on_gh.py --help` for usage information.
"""

import click
import json
import subprocess

ABBRV = {
    'xo': 'xobjects',
    'xd': 'xdeps',
    'xp': 'xpart',
    'xt': 'xtrack',
    'xf': 'xfields',
    'xm': 'xmask',
    'xc': 'xcoll',
    'xw': 'xwakes',
}

def make_flag(pkg):
    full_name = ABBRV[pkg].title()

    return click.option(
        f'--{pkg}',
        default='xsuite:main',
        help=f'{full_name} ref.',
        show_default=True,
    )

@click.command()
@make_flag('xo')
@make_flag('xd')
@make_flag('xp')
@make_flag('xt')
@make_flag('xf')
@make_flag('xm')
@make_flag('xc')
@make_flag('xw')
@click.option(
    '--platform',
    default='self-hosted',
    help='A label of the runner that should take this job.',
    show_default=True,
)
@click.option(
    '--ctx',
    default='cpu,omp,cuda,cl',
    help=(
        'Contexts on which to run tests. Comma separated '
        'list of cpu, omp, cuda, or cl with arguments '
        'passed after the colon, e.g. cpu:auto (equivalent '
        'to omp).'
    ),
    show_default=True,
)
@click.option(
    '--suites',
    default='xo,xd,xp,xt,xf,xc,xw',
    help='Test suites to run.',
    show_default=True,
)
@click.option(
    '--wf',
    default='manual_test_sh.yaml',
    help='The workflow file to use.',
    show_default=True,
)
@click.option(
    '--branch',
    default='main',
    help='The branch of the workflow.',
    show_default=True,
)
@click.option(
    '--pytest-opts',
    default='',
    help='Commandline options to pass to pytest.',
    show_default=True,
)
def run(xo, xd, xp, xt, xf, xm, xc, xw, platform, ctx, suites, wf, branch, pytest_opts):
    """Schedule a test run of Xsuite on a self-hosted runner.

    Example:

    python run_on_gh.py --suites xf,xd --platform pcbe-abp-gpu001-1 --xo xsuite:release/0.2.10 --xf xd:release/v0.5.0 --xt xsuite:release/0.46.2 --xf xsuite:release/0.14.1 --ctx cuda:3
    """
    try:
        subprocess.run(('gh', '--version'))
    except FileNotFoundError:
        print(__doc__)
        return 1

    ctx = ctx.replace('cl', 'ContextPyopencl')
    ctx = ctx.replace('cuda', 'ContextCupy')
    ctx = ctx.replace('cpu', 'ContextCpu')
    ctx = ctx.replace('omp', 'ContextCpu:auto')
    fmt_contexts = ctx.split(',')

    fmt_suites = [ABBRV[x.strip()] for x in suites.split(',')]

    parameters = {
        'locations' :json.dumps({
            'xobjects_location': xo,
            'xdeps_location': xd,
            'xpart_location': xp,
            'xtrack_location': xt,
            'xfields_location': xf,
            'xmask_location': xm,
            'xcoll_location': xc,
            'xwakes_location': xw,
        }) ,
        'pytest_options': pytest_opts,
        'test_contexts': ';'.join(fmt_contexts),
        'platform': platform,
        'suites': json.dumps(fmt_suites),
    }

    print('Scheduling')
    print(json.dumps(parameters, indent=2))

    workflow_command = f'gh workflow run {wf} --json --ref {branch}'

    res = subprocess.run(
        workflow_command.split(),
        input=json.dumps(parameters).encode(),
    )
    res.check_returncode()


if __name__ == '__main__':
    run()
